/*
   ancillary.h
   Ancillary routines for matrix diagonalization
   this file is part of the Diag library
   last modified 24 Aug 15 th
*/


#include "diag-c.h"

#define MatMul(ni, nj, nk, M, op) do { \
  int i, j, k; \
  for( i = 0; i < ni; ++i ) \
    for( j = 0; j < nj; ++j ) { \
      ComplexType s = 0; \
      for( k = 0; k < nk; ++k ) s += op; \
      Element(M,i,j) = s; \
    } \
} while( 0 )


/* IdentityMatrix
   set the n-by-n matrix A to the identity matrix. */

static inline void IdentityMatrix(cint n, ComplexType *A, cint ldA)
{
  int i;
  for( i = 0; i < n; ++i ) {
    memset(&A(i,0), 0, n*sizeof(ComplexType));
    A(i,i) = 1;
  }
}


/* ReverseHEigen
   recover the original matrix A from the eigenvalues and
   the transformation matrix U by forming A = U^+ diag(d) U. */

static inline void ReverseHEigen(cint n, ComplexType *A, cint ldA,
  cRealType *d, cComplexType *U, cint ldU)
{
#if UCOLS
  MatMul(n, n, n, A, d[k]*U(i,k)*Conjugate(U(j,k)));
#else
  MatMul(n, n, n, A, d[k]*Conjugate(U(k,i))*U(k,j));
#endif
}


/* ReverseSEigen
   recover the original matrix A from the (complex) eigenvalues
   and the transformation matrix U by forming A = U^T diag(d) U
   (or U diag(d) U^T if UCOLS=1). */

static inline void ReverseSEigen(cint n, ComplexType *A, cint ldA,
  cComplexType *d, cComplexType *U, cint ldU)
{
  MatMul(n, n, n, A, d[k]*UL(k,i)*UL(k,j));
}


/* ReverseTakagi
   recover the original matrix A from the eigenvalues and
   the transformation matrix U by forming A = U^T diag(d) U
   (or U^T diag(d) U if UCOLS=1). */

static inline void ReverseTakagi(cint n, ComplexType *A, cint ldA,
  cRealType *d, cComplexType *U, cint ldU)
{
  MatMul(n, n, n, A, d[k]*UL(k,i)*UL(k,j));
}


/* ReverseSVD
   recover the original matrix A from the singular values and
   the left and right transformation matrices V and W by
   forming A = V^T diag(d) W (or A = V d W^T if UCOLS=1). */

static inline void ReverseSVD(cint m, cint n, ComplexType *A, cint ldA,
  cRealType *d, cComplexType *V, cint ldV, cComplexType *W, cint ldW)
{
  cint nm = imin(m, n);
  MatMul(m, n, nm, A, d[k]*VL(k,i)*WL(k,j));
}


/* ApplyHEigen
   transform the n-by-n matrix A by a similarity transformation U,
   i.e. form U A U^+ (or U^+ A U if UCOLS=1).
   The original matrix A is overwritten with the result. */

static inline void ApplyHEigen(cint n, ComplexType *A, cint ldA,
  cComplexType *U, cint ldU)
{
  ComplexType T[n*n];
  cint ldT = n;
#if UCOLS
  MatMul(n, n, n, T, A(i,k)*U(k,j));
  MatMul(n, n, n, A, Conjugate(U(k,i))*Element(T,k,j));
#else
  MatMul(n, n, n, T, A(i,k)*Conjugate(U(j,k)));
  MatMul(n, n, n, A, U(i,k)*Element(T,k,j));
#endif
}


/* ApplyTakagi
   transform the n-by-n matrix A by a unitary transformation U,
   i.e. form U^* A U^+ (or U^+ A U^* if UCOLS=1).
   The original matrix A is overwritten with the result. */

static inline void ApplyTakagi(cint n, ComplexType *A, cint ldA,
  cComplexType *U, cint ldU)
{
  ComplexType T[n*n];
  cint ldT = n;
  MatMul(n, n, n, T, A(i,k)*Conjugate(UL(j,k)));
  MatMul(n, n, n, A, Conjugate(UL(i,k))*Element(T,k,j));
}


/* ApplySVD
   transform the n-by-n matrix A by the left and right transformation
   matrices V and W, i.e. form V^* A W^+.
   The original matrix A is overwritten with the result. */

static inline void ApplySVD(cint m, cint n, ComplexType *A, cint ldA,
  cComplexType *V, cint ldV, cComplexType *W, cint ldW)
{
  cint nm = imin(m, n);
  ComplexType T[nm*nm];
  cint ldT = nm;
  MatMul(nm, n, m, T, Conjugate(VL(i,k))*A(k,j));
  MatMul(nm, nm, n, A, Element(T,i,k)*Conjugate(WL(j,k)));
}

